/*
 * Copyright (c) 2017 Samsung Electronics Co., Ltd. All rights reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/*global window, console, document */

/**
 * Application UI module.
 * It is responsible for managing user interface.
 *
 * @module app.ui
 * @requires {@link app.model}
 * @requires {@link app.ui.templates}
 * @namespace app.ui
 * @memberof app
 */

// make sure that "app" namespace is created
window.app = window.app || {};

(function defineAppUi(app) {
    'use strict';

    /**
     * Remove all message.
     *
     * @memberof app.ui
     * @private
     * @const {string}
     */
    var REMOVE_ALL_CONFIRM_MSG =
            'Are you sure you want to remove all preferences?',

        /**
         * Remove single preference message.
         *
         * @memberof app.ui
         * @private
         * @const {string}
         */
        REMOVE_ONE_CONFIRM_MSG =
            'Are you sure you want to remove this preference?',

        /**
         * Offset in pixels required to trigger list item swipe action.
         *
         * @memberof app.ui
         * @private
         * @const {string}
         */
        SWIPE_THRESHOLD = 50,

        /**
         * CSS class used for list item delete mode.
         *
         * @memberof app.ui
         * @private
         * @const {string}
         */
        DELETE_MODE_CLASS = 'delete-mode',

        /**
         * Reference to HTML element containing the list of stored preferences.
         *
         * @memberof app.ui
         * @private
         * @type {HTMLElement}
         */
        listElement = null,

        /**
         * Reference to HTML section element containing preferences list.
         *
         * @memberof app.ui
         * @private
         * @type {HTMLElement}
         */
        listSectionElement = null,

        /**
         * Reference to HTML input which allows user to provide the key name
         * for the preference.
         *
         * @memberof app.ui
         * @private
         * @type {HTMLElement}
         */
        keyInputElement = null,

        /**
         * Reference to HTML input which allows user to provide the value
         * for the preference.
         *
         * @memberof app.ui
         * @private
         * @type {HTMLElement}
         */
        valueInputElement = null,

        /**
         * Reference to HTML element containing radio buttons which allow user
         * to change the preference type.
         *
         * @memberof app.ui
         * @private
         * @type {HTMLElement}
         */
        typeFieldsetElement = null,

        /**
         * Reference to pop-up overlay HTML element.
         *
         * @memberof app.ui
         * @private
         * @type {HTMLElement}
         */
        popupOverlayElement = null,

        /**
         * Reference to pop-up message HTML element.
         *
         * @memberof app.ui
         * @private
         * @type {HTMLElement}
         */
        popupMsgElement = null,

        /**
         * Reference to the "Update" button element.
         *
         * @memberof app.ui
         * @private
         * @type {HTMLElement}
         */
        updateBtn = null,

        /**
         * Reference to the "Remove all" button element.
         *
         * @memberof app.ui
         * @private
         * @type {HTMLElement}
         */
        removeAllBtn = null,

        /**
         * Reference to the pop-up confirm button.
         *
         * @memberof app.ui
         * @private
         * @type {HTMLElement}
         */
        popupConfirmBtn = null,

        /**
         * Reference to the pop-up cancel button.
         *
         * @memberof app.ui
         * @private
         * @type {HTMLElement}
         */
        popupCancelBtn = null,

        /**
         * Indicates which preference key will be removed
         * after confirmation.
         *
         * @memberof app.ui
         * @private
         * @type {string}
         */
        keyToRemove = '',

        /**
         * Preference item swipe data.
         *
         * @memberof app.ui
         * @private
         * @type {object}
         */
        prefItemSwipeData = {
            object: null,
            startX: 0,
            currentX: 0,
            offsetX: 0,
            enabled: false
        };

    /**
     * Handles back action.
     *
     * Returns true if action was handled, false otherwise (owner module should
     * handle it).
     *
     * @memberof app.ui
     * @public
     * @returns {boolean}
     */
    function handleBackAction() {
        return false;
    }

    /**
     * Returns type name of specified preference value.
     *
     * @memberof app.ui
     * @private
     * @param {PreferenceValueType} value
     * @returns {string}
     */
    function getPreferenceValueType(value) {
        switch (typeof value) {
        case 'boolean':
            return 'Bool';
        case 'string':
            return 'String';
        case 'number':
            return app.common.isInteger(value) ? 'Integer' : 'Double';
        default:
            return 'Other';
        }
    }

    /**
     * Updates the list content using specified preferences data.
     *
     * @memberof app.ui
     * @private
     * @param {PreferenceData[]} preferences
     */
    function updateList(preferences) {
        var i = 0,
            itemsCount = preferences.length,
            preference = null,
            newContent = '';

        if (itemsCount === 0) {
            listSectionElement.classList.add('empty');
        } else {
            for (i = 0; i < itemsCount; i += 1) {
                preference = preferences[i];

                newContent += app.ui.templates.getListElement({
                    id: preference.key,
                    key: preference.key,
                    value: preference.value.toString(),
                    type: getPreferenceValueType(preference.value)
                });
            }
            listSectionElement.classList.remove('empty');
        }

        listElement.innerHTML = newContent;
    }

    /**
     * Called when preferences are received.
     * Updates the list content and related buttons state.
     *
     * @memberof app.ui
     * @private
     * @param {PreferenceData[]} preferences
     */
    function onPreferencesGetSuccess(preferences) {
        updateList(preferences);
        removeAllBtn.disabled = preferences.length === 0;
    }

    /**
     * Handles preferences get error.
     *
     * @memberof app.ui
     * @private
     * @param {Error} error
     */
    function onPreferencesGetError(error) {
        console.error('Error occurred during obtaining preferences data',
            error);
        window.alert('Error occurred during obtaining preferences data.');
    }

    /**
     * Triggers refreshing the list of preferences.
     *
     * @memberof app.ui
     * @private
     */
    function refresh() {
        app.model.getPreferences(
            onPreferencesGetSuccess, onPreferencesGetError);
    }

    /**
     * Returns true if "Value" form field is valid, false otherwise.
     *
     * @memberof app.ui
     * @private
     * @returns {boolean}
     */
    function isValueFormFieldValid() {
        var selectedTypeRadio = typeFieldsetElement.querySelector(':checked'),
            value = valueInputElement.value.trim();

        switch (selectedTypeRadio.value) {
        case 'Bool':
            return ['false', 'true'].indexOf(value) !== -1;
        case 'Integer':
            return value.length > 0 && !isNaN(value) &&
                value.indexOf('.') === -1;
        case 'Double':
            return !isNaN(value) && value.indexOf('.') !== -1;
        case 'String':
            return value.length > 0;
        default:
            return false;
        }
    }

    /**
     * Activates update button if user has typed key and value.
     *
     * @memberof app.ui
     * @private
     */
    function checkFormCompletion() {
        var key = keyInputElement.value.trim(),
            value = valueInputElement.value.trim(),
            valid = false;

        valid = key.length > 0 && value.length > 0;

        updateBtn.disabled = !valid;
    }

    /**
     * Resets the form fields to default values.
     * After the operation, form validation is triggered.
     *
     * @membeof app.ui
     * @private
     */
    function resetForm() {
        keyInputElement.value = '';
        valueInputElement.value = '';
        typeFieldsetElement.querySelector('#bool-type-radio').checked = true;
        checkFormCompletion();
    }

    /**
     * Shows overlay with delete pop-up.
     *
     * @memberof app.ui
     * @private
     */
    function showPopup() {
        document.body.classList.add('active-popup');
    }

    /**
     * Hides overlay with delete pop-up.
     *
     * @memberof app.ui
     * @private
     */
    function closePopup() {
        document.body.classList.remove('active-popup');
    }

    /**
     * Handles value change in "Key name" form field.
     * Triggers form validation.
     *
     * @memberof app.ui
     * @private
     */
    function onKeyFormFieldChanged() {
        checkFormCompletion();
    }

    /**
     * Handles value change in "Value" form field.
     * Triggers form validation.
     *
     * @memberof app.ui
     * @private
     */
    function onValueFormFieldChanged() {
        checkFormCompletion();
    }

    /**
     * Handles change of selected type radio input.
     * Clears the value input and triggers form validation.
     *
     * @memberof app.ui
     * @private
     */
    function onTypeFormFieldChanged() {
        checkFormCompletion();
    }

    /**
     * Handles click on the "Update" button.
     *
     * @memberof app.ui
     * @private
     */
    function onUpdateBtnClick() {
        var selectedTypeRadio = typeFieldsetElement.querySelector(':checked'),
            value = valueInputElement.value.trim(),
            key = keyInputElement.value.trim(),
            updatedSuccessfully = false,
            validValue = isValueFormFieldValid();

        if (!validValue) {
            window.alert('Incorrect value.');
            return;
        }

        switch (selectedTypeRadio.value) {
        case 'Bool':
            value = value === 'true';
            break;
        case 'Integer':
            value = parseInt(value, 10);
            break;
        case 'Double':
            value = parseFloat(value);
            break;
        case 'String':
            // value is already trimmed, do nothing
            break;
        default:
            window.alert('Wrong data type.');
            return;
        }

        updatedSuccessfully = app.model.setPreferenceValue(key, value);

        if (updatedSuccessfully) {
            resetForm();
            refresh();
        } else {
            window.alert('Error occurred during updating preference.');
        }
    }

    /**
     * Handles click on the "Remove all" button.
     *
     * @memberof app.ui
     * @private
     */
    function onRemoveAllBtnClick() {
        popupMsgElement.innerText = REMOVE_ALL_CONFIRM_MSG;
        keyToRemove = 'ALL';
        showPopup();
    }

    /**
     * Handles click on the remove icon.
     *
     * @memberof app.ui
     * @private
     * @param {Event} e Click event.
     */
    function onRemoveItemBtnClick(e) {
        popupMsgElement.innerText = REMOVE_ONE_CONFIRM_MSG;
        keyToRemove = e.srcElement.parentElement.dataset.id;
        showPopup();
    }

    /**
     * Handles delete confirmation.
     * Depending on the state of 'keyToRemove' this function
     * will trigger remove all or remove specified preference action.
     *
     * @memberof app.ui
     * @private
     */
    function onPopupConfirmBtnClick() {
        var removedSuccessfully;

        if (keyToRemove !== 'ALL') {
            removedSuccessfully = app.model.removePreference(keyToRemove);
        } else {
            removedSuccessfully = app.model.removePreferences();
        }

        if (removedSuccessfully) {
            refresh();
        } else {
            window.alert('Error occurred during removing preferences.');
        }

        completeDeleteAction();
    }

    /**
     * Clears the state of elements used in delete process
     * and closes the pop-up overlay.
     * It should be called after remove action or in case of
     * cancelation delete action.
     *
     * @memberof app.ui
     * @private
     */
    function completeDeleteAction() {
        popupMsgElement.innerText = '';
        keyToRemove = '';
        closePopup();
    }

    /**
     * Handles click on the list element.
     * Determines clicked list item to populate the preference form
     * with corresponding data.
     *
     * @memberof app.ui
     * @private
     * @param {MouseEvent} event
     */
    function onListClick(event) {
        var targetListElement = app.common.closestElement(event.target,
                'li.list-item'),
            preferenceValue = null;

        if (event.srcElement.tagName === 'BUTTON') {
            onRemoveItemBtnClick(event);
            return;
        }

        if (!targetListElement) {
            return;
        }

        preferenceValue = app.model.getPreferenceValue(
            targetListElement.dataset.id);

        if (preferenceValue === null) {
            return;
        }

        keyInputElement.value = targetListElement.dataset.id;
        valueInputElement.value = preferenceValue.toString();
        typeFieldsetElement
            .querySelector(
                'input[value="' + getPreferenceValueType(preferenceValue) + '"]'
            ).checked = true;

        checkFormCompletion();
    }

    /**
     * Handles touchstart event on preference list item.
     * Sets initial swipe coordinates.
     *
     * @memberof app.ui
     * @private
     * @param {TouchEvent} e
     */
    function onTouchStartPrefList(e) {
        if (prefItemSwipeData.object) {
            return;
        }

        prefItemSwipeData.enabled = true;
        prefItemSwipeData.object = app.common.closestElement(e.target,
            'li.list-item');

        prefItemSwipeData.startX = e.touches[0].pageX;
        prefItemSwipeData.currentX = e.touches[0].pageX;

    }

    /**
     * Handles touchmove event on preference list item.
     * Updates current X position and offset.
     *
     * @memberof app.ui
     * @private
     * @param {TouchEvent} e
     */
    function onTouchMovePrefList(e) {
        prefItemSwipeData.currentX = e.touches[0].pageX;

        prefItemSwipeData.offsetX =
            prefItemSwipeData.currentX -
            prefItemSwipeData.startX;
    }

    /**
     * Handles touchend event on preference list item.
     *
     * Toggles visibility of delete icon depending on offset value.
     * Only one list item may have visible delete icon at once.
     *
     * @memberof app.ui
     * @private
     */
    function onTouchEndPrefList() {
        var offset = prefItemSwipeData.offsetX,
            listItem = prefItemSwipeData.object,
            prevSwipedItem = null;

        if (offset > SWIPE_THRESHOLD) {
            listItem.classList.remove(DELETE_MODE_CLASS);
        } else if (offset < -SWIPE_THRESHOLD) {
            prevSwipedItem =
                listElement.querySelector('.' + DELETE_MODE_CLASS);

            if (prevSwipedItem) {
                prevSwipedItem.classList.remove(DELETE_MODE_CLASS);
            }
            listItem.classList.add(DELETE_MODE_CLASS);
        }

        prefItemSwipeData.object = null;
    }

    /**
     * Binds HTML elements to variables to improve access time.
     *
     * @memberof app.ui
     * @private
     */
    function bindElements() {
        listSectionElement = document.getElementById('list');
        listElement = document.querySelector('#list ul');
        keyInputElement = document.getElementById('key-name-input');
        valueInputElement = document.getElementById('value-input');
        typeFieldsetElement = document.getElementById('type-fieldset');
        updateBtn = document.getElementById('update-btn');
        removeAllBtn = document.getElementById('remove-all-btn');
        popupOverlayElement = document.getElementById('popup-overlay');
        popupMsgElement = document.getElementById('popup-msg');
        popupConfirmBtn = popupOverlayElement.querySelector('#confirm-btn');
        popupCancelBtn = popupOverlayElement.querySelector('#cancel-btn');
    }

    /**
     * Registers event listeners for the module.
     *
     * @memberof app.ui
     * @private
     */
    function bindEvents() {
        keyInputElement.addEventListener('input', onKeyFormFieldChanged);
        valueInputElement.addEventListener('input', onValueFormFieldChanged);
        typeFieldsetElement.addEventListener('change', onTypeFormFieldChanged);
        updateBtn.addEventListener('click', onUpdateBtnClick);
        removeAllBtn.addEventListener('click', onRemoveAllBtnClick);
        popupConfirmBtn.addEventListener('click', onPopupConfirmBtnClick);
        popupCancelBtn.addEventListener('click', completeDeleteAction);

        listElement.addEventListener('click', onListClick);
        listElement.addEventListener('touchstart', onTouchStartPrefList);
        listElement.addEventListener('touchmove', onTouchMovePrefList);
        listElement.addEventListener('touchend', onTouchEndPrefList);
    }

    /**
     * Starts the module by loading the data to the list.
     *
     * @memberof app.ui
     * @public
     */
    function run() {
        refresh();
    }

    /**
     * Initializes the module.
     *
     * @memberof app.ui
     * @public
     */
    function init() {
        bindElements();
        bindEvents();
    }

    app.ui = {
        init: init,
        handleBackAction: handleBackAction,
        run: run
    };

})(window.app);
